"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
from gfutilities.puls.pulsedata import decode_all_steps, generate_linear_puls

__all__ = ['decode_all_steps', 'generate_linear_puls']
